#include "packetsource.h"

// constructor for offline traces
PacketSource::PacketSource()
{
    this->PacketSrc = 0;
}

// Destructor
PacketSource::~PacketSource()
{
    this->CloseSource();
}

std::vector<Device> PacketSource::GetDeviceList()
{
    char errbuf[PCAP_ERRBUF_SIZE];
    pcap_if_t *alldevs;
    pcap_if_t *d;

    std::vector<Device> Devices;

    // Retrieve the device list from the local machine
    if (pcap_findalldevs(&alldevs, errbuf) == -1)
        throw Dev_Find_Fail(errbuf);

    // Build the vector
    for(d= alldevs; d != NULL; d= d->next)
        if (d->description)
            Devices.push_back(Device(d->name,d->description));
        else
            Devices.push_back(Device(d->name,"(No description available)"));

    pcap_freealldevs(alldevs);

    return Devices;
}


// constructor for offline traces
void PacketSource::OpenOfflineSource(const char *FileName)
{
    char errbuf[PCAP_ERRBUF_SIZE];

    if(this->PacketSrc != 0)
        throw Src_Already_Open();

    this->PacketSrc = pcap_open_offline(FileName, errbuf);

    std::string error_msg = errbuf;

    if(this->PacketSrc == 0)
        throw Dev_Open_Offline_Fail(error_msg);

    /* Check the link layer. Only Ethernet is supported. */
	if(pcap_datalink(this->PacketSrc) != DLT_EN10MB)
	{
	    this->CloseSource();
        throw Not_Ethernet_Source();
	}
}

// constructor for live captures
void PacketSource::OpenOnlineSource(const char *device, const int &MaxSize, const int &Timeout)
{
    char errbuf[PCAP_ERRBUF_SIZE];

    if(this->PacketSrc != 0)
        throw Src_Already_Open();

    /* open the device for sniffing.

       pcap_t *pcap_open_live(char *device,int snaplen, int prmisc,int to_ms,
       char *ebuf)

       snaplen - maximum size of packets to capture in bytes
       promisc - set card in promiscuous mode?
       to_ms   - time to wait for packets in miliseconds before read
       times out
       errbuf  - if something happens, place error string here

       Note if you change "prmisc" param to anything other than zero, you will
       get all packets your device sees, whether they are intendeed for you or
       not!! Be sure you know the rules of the network you are running on
       before you set your card in promiscuous mode!!     */

    strcpy(errbuf, "");
    this->PacketSrc = pcap_open_live(device,MaxSize,0,Timeout,errbuf);

    std::string error_msg = errbuf;

    if(this->PacketSrc == 0)
        throw Dev_Open_Live_Fail(error_msg);
    else if (error_msg.length() != 0)
        throw Dev_Open_Live_Warning(error_msg);

    /* Check the link layer. Only Ethernet is supported. */
	if(pcap_datalink(this->PacketSrc) != DLT_EN10MB)
	{
	    this->CloseSource();
        throw Not_Ethernet_Source();
	}
}

void PacketSource::CloseSource()
{
    if ((this->PacketSrc) != 0)
    {
        pcap_close(this->PacketSrc);
        this->PacketSrc = 0;
    }
}

// returns the next Ethernet frame in a Packet, EthPacket, IPPacket, UDPPacket, or TCPPacket object, depending on the highest protocol.
Packet * PacketSource::NextPacket()
{
    if (!this->IsOpen())
        throw Src_Not_Open();

    const struct pcap_pkthdr *pkt_header;// = new struct pcap_pkthdr;
    const u_char *pkt_data;// = new u_char;
    int iRes;

    struct ether_header *et_hd;
    struct ip *ip_hd;

    iRes = pcap_next_ex(this->PacketSrc, const_cast<struct pcap_pkthdr **>(&pkt_header), &pkt_data);

    switch (iRes)
    {
        case 0:
            throw Timeout_Expired();
        case -1:
        {
            std::string error_msg = pcap_geterr(this->PacketSrc);
            throw Packet_Reading_Error(error_msg);
        }
        case -2:
            throw No_More_Packets();
    }

    /* lets start with the ether header... */
    et_hd = (struct ether_header *) pkt_data;

    /* check to see if we have an ip packet */
    if (ntohs (et_hd->ether_type) == ETHERTYPE_IP)
    {
        ip_hd = (struct ip *) (pkt_data + ETHER_HDRLEN);

        if ((ip_hd->ip_p) == IPPROTO_TCP)
            return new TCPPacket(pkt_header, pkt_data);

        else if ((ip_hd->ip_p) == IPPROTO_UDP)
            return new UDPPacket(pkt_header, pkt_data);

        else
            return new IPPacket(pkt_header, pkt_data);
    }
    else
       return new EthPacket(pkt_header, pkt_data);

    return 0;
}

bool PacketSource::IsOpen() const
{
    if (this->PacketSrc != 0)
        return true;
    else
        return false;
}
