#include "entropycontainer.h"

template<class V> EntropyContainer<V>::EntropyContainer() {

}

template<class V> EntropyContainer<V>::EntropyContainer(const int &WindowSize)
{
    init(WindowSize);
}

template<class V> EntropyContainer<V>::~EntropyContainer()
{
    this->dispose();
}

template<class V> void EntropyContainer<V>::init(const int &WindowSize)
{
    this->CurrentEntropy = 0;
    this->Total = 0;
    this->MinEntropy = 0;
    this->MaxEntropy = 0;
    this->ExpMean = 0;

    this->EntropyDump = new MeanVarianceDump();
    this->MaxDump = new MeanVarianceDump();
    this->MinDump = new MeanVarianceDump();
    this->ValueDump = new MeanVarianceDump();
    this->Win = new SlidingWindow<V>(WindowSize);

    const int numStats = 16;
    this->aStats = new double[numStats];
}

template<> void EntropyContainer<std::string>::init(const int &WindowSize)
{
    this->CurrentEntropy = 0;
    this->Total = 0;
    this->MinEntropy = 0;
    this->MaxEntropy = 0;
    this->ExpMean = 0;

    this->EntropyDump = new MeanVarianceDump();
    this->MaxDump = 0;
    this->MinDump = 0;
    this->ValueDump = 0;
    this->Win = new SlidingWindow<std::string>(WindowSize);

    const int numStats = 6;
    this->aStats = new double[numStats];
}

template<class V> void EntropyContainer<V>::Clear()
{
    this->CurrentEntropy = 0;
    this->Total = 0;
    this->MinEntropy = 0;
    this->MaxEntropy = 0;
    this->ExpMean = 0;

    this->dispose();
}

template<class V> const int EntropyContainer<V>::WindowSize() const
{
    return this->Win->WinSize();
}

template<class V> void EntropyContainer<V>::dispose()
{
	delete this->Win;
	delete this->EntropyDump;
    delete this->MaxDump;
    delete this->MinDump;
    delete this->ValueDump;
    delete [] this->aStats;
}

template<class V> const double EntropyContainer<V>::EntropyVal() const
{
	if (this->Win->IsFilled())
	{
		// ---------------------------------------------------------------------------------------------//
		// The operations used to implement the "on the fly" entropy calculation sometimes generate		//
		// extremely low values, near zero. Since the precision of the double type is not sufficient	//
		// to represent such values, they are rounded by the intermediate mathematical operations.		//
		// Because of that, when the entropy value decreases to zero, the CurrentEntropy double			//
		// variable contains a negative value, extremely close to zero. When this value is printed,		//
		// a negative value like "-0.000000" is obtained.												//
		if(this->CurrentEntropy < 0.0)
			return 0;
		// ---------------------------------------------------------------------------------------------//

		return this->CurrentEntropy;
	}
    return 0;
}

template<class V> const double EntropyContainer<V>::EntropyMean() const
{
	if (this->Win->IsFilled())
	{
		// ---------------------------------------------------------------------------------------------//
		// Since the 'Mean' value depends on the 'Entropy' value, a similar operation to the one made	//
		// in the 'EntropyVal' method has to be made here.												//
		if(this->EntropyDump->Mean() < 0.0)
			return 0;
		// ---------------------------------------------------------------------------------------------//

		return this->EntropyDump->Mean();
	}
    return 0;
}

template<class V> const double EntropyContainer<V>::EntropyVariance() const
{
	if (this->Win->IsFilled())
	{
		// ---------------------------------------------------------------------------------------------//
		// Since the 'Variance' value depends on the 'Entropy' value, a similar operation to the one	//
		// made in the 'EntropyVal' method has to be made here.											//
		if(this->EntropyDump->Variance() < 0.0)
			return 0;
		// ---------------------------------------------------------------------------------------------//

		return this->EntropyDump->Variance();
	}
    return 0;
}

template<class V> const double EntropyContainer<V>::EntropyMax() const
{
	return this->MaxEntropy;
}

template<class V> const double EntropyContainer<V>::EntropyMin() const
{
	return this->MinEntropy;
}

template<class V> const double EntropyContainer<V>::ValueMean() const
{
	return this->ValueDump->Mean();
}

template<class V> const double EntropyContainer<V>::ValueExpMean() const
{
	return this->ExpMean;
}

template<class V> const double EntropyContainer<V>::WinMean() const
{
	return this->Win->Mean();
}

template<class V> const double EntropyContainer<V>::ValueVariance() const
{
	return this->ValueDump->Variance();
}

template<class V> const double EntropyContainer<V>::MaxMean() const
{
	return this->MaxDump->Mean();
}

template<class V> const double EntropyContainer<V>::MinMean() const
{
	return this->MinDump->Mean();
}

template<class V> const double EntropyContainer<V>::MaxVariance() const
{
	return this->MaxDump->Variance();
}

template<class V> const double EntropyContainer<V>::MinVariance() const
{
	return this->MinDump->Variance();
}

template<class V> void EntropyContainer<V>::AddValue(const V &value)
{
	const int iWin = this->Win->WinSize();

    // If the window is already filled:
    if (this->Win->IsFilled())
    {
        V Elem = this->Win->OldestElem();

        if (Elem != value)
        {
        	int const ElemOccur = this->Win->Count(Elem);
            // Removes the influence of the oldest element in the entropy value
        	this->CurrentEntropy += (ElemOccur/((double) (iWin))) * ((double) (log(ElemOccur/((double) (iWin)))));
        	if (ElemOccur > 1)
        		this->CurrentEntropy -= ((ElemOccur-1)/((double) (iWin))) * ((double) (log((ElemOccur-1)/((double) (iWin)))));

            int const valueOccur = this->Win->Count(value);
        	// If this is not the first occurrence of the element value:
            if (valueOccur > 0)
                // Removes its influence in the entropy value
            	this->CurrentEntropy += (valueOccur/((double) (iWin))) * ((double) (log(valueOccur/((double) (iWin)))));

            // Updates the entropy value for the new element
            this->CurrentEntropy -= ((valueOccur+1)/((double)(iWin))) * ((double) (log((valueOccur+1)/((double) (iWin)))));
        }
    }
    else {
        int const valueOccur = this->Win->Count(value);

		// If this is not the first occurrence of the element value:
		if (valueOccur > 0)
			// Removes its influence in the entropy value
			this->CurrentEntropy += (valueOccur/((double) (iWin))) * ((double) (log(valueOccur/((double) (iWin)))));

		// Updates the entropy value for the new element
		this->CurrentEntropy -= ((valueOccur+1)/((double)(iWin))) * ((double) (log((valueOccur+1)/((double) (iWin)))));
    }

    this->Win->AddElement(value);

    (this->Total)++;

    this->UpdateLimits();
    this->UpdateMean();
    this->UpdateExpMean(value);
}

template<class V> const long EntropyContainer<V>::TotalAnalyzedItems() const {
	return this->Total;
}

template<class V> void EntropyContainer<V>::UpdateLimits()
{
	if (this->Win->IsFilled())
	{
		if (this->Total == this->Win->WinSize())
		{
			this->MaxEntropy = this->CurrentEntropy;
			this->MinEntropy = this->CurrentEntropy;
		}
		else
		{
			if (this->CurrentEntropy > this->MaxEntropy)
				this->MaxEntropy = this->CurrentEntropy;
			else if (this->CurrentEntropy < this->MinEntropy)
				this->MinEntropy = this->CurrentEntropy;
		}
	}
}

template<class V> void EntropyContainer<V>::UpdateMean()
{
	if (this->Win->IsFilled())
	{
		this->EntropyDump->AddValue(this->EntropyVal());
		this->MaxDump->AddValue(this->Win->Max());
		this->MinDump->AddValue(this->Win->Min());
	}

	this->ValueDump->AddValue(this->Win->NewestElem());
}

template<> void EntropyContainer<std::string>::UpdateMean()
{
	if (this->Win->IsFilled())
		this->EntropyDump->AddValue(this->EntropyVal());
}

template<class V> void EntropyContainer<V>::UpdateExpMean(const V &Val)
{
	if (this->ExpMean == 0)
		this->ExpMean = Val;
	else
		this->ExpMean = (this->ExpMean * (this->Win->WinSize() - 1) + Val) / this->Win->WinSize();
}

template<> void EntropyContainer<std::string>::UpdateExpMean(const std::string &Val)
{

}

template<class V> const V EntropyContainer<V>::Max() const {
	return this->Win->Max();
}

template<class V> const V EntropyContainer<V>::Min() const {
	return this->Win->Min();
}

template<> double * EntropyContainer<std::string>::GetStats() {
	this->aStats[0] = this->EntropyVal();
	this->aStats[1] = this->EntropyMean();
	this->aStats[2] = this->EntropyVariance();
	this->aStats[3] = this->MaxEntropy;
	this->aStats[4] = this->MinEntropy;
	this->aStats[5] = this->Total;

	return this->aStats;
}

template<class V> double * EntropyContainer<V>::GetStats() {
	this->aStats[0] = this->EntropyVal();
	this->aStats[1] = this->EntropyMean();
	this->aStats[2] = this->EntropyVariance();
	this->aStats[3] = this->MaxEntropy;
	this->aStats[4] = this->MinEntropy;
	this->aStats[5] = this->Total;
	this->aStats[6] = this->ValueMean();
	this->aStats[7] = this->ExpMean;
	this->aStats[8] = this->WinMean();
	this->aStats[9] = this->ValueVariance();
	this->aStats[10] = this->Max();
	this->aStats[11] = this->MaxMean();
	this->aStats[12] = this->MaxVariance();
	this->aStats[13] = this->Min();
	this->aStats[14] = this->MinMean();
	this->aStats[15] = this->MinVariance();

	return this->aStats;
}

template<class V> double * EntropyContainer<V>::GetStatsClassification() {
	this->aStats[0] = this->EntropyVal();
	this->aStats[5] = this->Total;
	this->aStats[6] = this->ValueMean();

	return this->aStats;
}

template class EntropyContainer<int>;
template class EntropyContainer<double>;
template class EntropyContainer<std::string>;
