#include "flowanalyzer.h"

template<class V> FlowAnalyzer<V>::FlowAnalyzer(const int &WindowSize)
{
    init(WindowSize, std::string(), 0);
}

template<class V> FlowAnalyzer<V>::FlowAnalyzer(const int &WindowSize, const std::string &LogsPath)
{
    init(WindowSize, LogsPath, 0);
}

template<class V> FlowAnalyzer<V>::FlowAnalyzer(const int &WindowSize, const std::string &LogsPath, ThresholdsSet * TSet)
{
    init(WindowSize, LogsPath, TSet);
}

template<class V> FlowAnalyzer<V>::~FlowAnalyzer()
{
    this->dispose();
}

template<class V> void FlowAnalyzer<V>::Clear()
{
    this->dispose();

    this->Flows.clear();
    this->NumPackets.clear();
    this->LastTS.clear();
    this->HalfClosedCSflows.clear();
    this->HalfClosedSCflows.clear();
    this->NumFlows.clear();
    this->LastValue.clear();

    this->Thresholds = 0;

    this->TotalTCPflows = 0;
    this->TotalTCPpackets = 0;
    this->TotalUDPflows = 0;
    this->TotalUDPpackets = 0;

    this->WinSize = 0;
}

template<class V> void FlowAnalyzer<V>::dispose()
{
    typename std::map<std::string, EntropyContainer<V> *>::iterator it;

    for(it = this->Flows.begin(); it != this->Flows.end(); it++)
        delete (it->second);
}

template<class V> void FlowAnalyzer<V>::init(const int &WindowSize, const std::string &Path, ThresholdsSet * TSet)
{
    this->TotalTCPflows = 0;
    this->TotalTCPpackets = 0;
    this->TotalUDPflows = 0;
    this->TotalUDPpackets = 0;
    this->WinSize = WindowSize;

    this->Thresholds = TSet;

    this->LogsFolder.InitFolder(Path);
}

template<class V> void FlowAnalyzer<V>::WriteEntropy(const std::string &Key, const double &TStamp, const V &Value, EntropyContainer<V>* Entropy)
{
    std::stringstream ss (std::stringstream::out);

	double dTimeElapsed;

    if (this->FirstTS.count(Key) == 0)
		this->FirstTS.insert(std::pair<std::string, double>(Key, TStamp));

    dTimeElapsed = TStamp - (this->FirstTS[Key]);

    if (this->Thresholds != 0)
    	ss << std::fixed << TStamp << "\t"  << std::fixed << std::setprecision(6) << dTimeElapsed << "\t" << Value << "\t" << Entropy->EntropyVal() << "\t" << Entropy->EntropyMean() << "\t" << Entropy->EntropyVariance() << "\t" << Entropy->EntropyMin() << "\t" << Entropy->EntropyMax() << "\t" << std::setprecision(2) << Entropy->ValueMean() << "\t" << Entropy->ValueVariance() << "\t" << Entropy->Min() << "\t" << Entropy->Max() << "\t" << Entropy->MinMean() << "\t" << Entropy->MaxMean() << "\t" << Entropy->MinVariance() << "\t" << Entropy->MaxVariance() << "\t" << this->Thresholds->GetDescription(Entropy->EntropyVal()) << std::endl;
    else
    	ss << std::fixed << TStamp << "\t"  << std::fixed << std::setprecision(6) << dTimeElapsed << "\t" << Value << "\t" << Entropy->EntropyVal() << "\t" << Entropy->EntropyMean() << "\t" << Entropy->EntropyVariance() << "\t" << Entropy->EntropyMin() << "\t" << Entropy->EntropyMax() << "\t" << std::setprecision(2) << Entropy->ValueMean() << "\t" << Entropy->ValueVariance() << "\t" << Entropy->Min() << "\t" << Entropy->Max() << "\t" << Entropy->MinMean() << "\t" << Entropy->MaxMean() << "\t" << Entropy->MinVariance() << "\t" << Entropy->MaxVariance() << std::endl;

    this->WriteExistentLog(Key, ss.str());
}

template<> void FlowAnalyzer<std::string>::WriteEntropy(const std::string &Key, const double &TStamp, const std::string &Value, EntropyContainer<std::string>* Entropy)
{
    std::stringstream ss (std::stringstream::out);

	double dTimeElapsed;

    if (this->FirstTS.count(Key) == 0)
		this->FirstTS.insert(std::pair<std::string, double>(Key, TStamp));

    dTimeElapsed = TStamp - (this->FirstTS[Key]);

    if (this->Thresholds != 0)
    	ss << std::fixed << TStamp << "\t"  << std::fixed << std::setprecision(6) << dTimeElapsed << "\t" << Value << "\t" << Entropy->EntropyVal() << "\t" << Entropy->EntropyMean() << "\t" << Entropy->EntropyVariance() << "\t" << Entropy->EntropyMin() << "\t" << Entropy->EntropyMax() << "\t" << this->Thresholds->GetDescription(Entropy->EntropyVal()) << std::endl;
    else
    	ss << std::fixed << TStamp << "\t"  << std::fixed << std::setprecision(6) << dTimeElapsed << "\t" << Value << "\t" << Entropy->EntropyVal() << "\t" << Entropy->EntropyMean() << "\t" << Entropy->EntropyVariance() << "\t" << Entropy->EntropyMin() << "\t" << Entropy->EntropyMax() << std::endl;

    this->WriteExistentLog(Key, ss.str());
}

template<class V> void FlowAnalyzer<V>::WriteNewLog(const std::string &Key, const std::string &Msg)
{
	std::stringstream ssNumFlows (std::stringstream::out);
    ssNumFlows << this->NumFlows[Key];

	const std::string fileName = Key + "#" + ssNumFlows.str() + ".log";

    this->LogsFolder.WriteNewFile(fileName, Msg);
}

template<class V> void FlowAnalyzer<V>::WriteExistentLog(const std::string &Key, const std::string &Msg)
{
    std::stringstream ssNumFlows (std::stringstream::out);
    ssNumFlows << this->NumFlows[Key];

	const std::string fileName = Key + "#" + ssNumFlows.str() + ".log";

    this->LogsFolder.WriteExistentFile(fileName, Msg);
}

template<class V> template<class D> double * FlowAnalyzer<V>::AddValueDif(const Packet *pPkt, const D &Value, int precision)
{
	std::string srcIP, dstIP, Proto;
	int sPort, dPort, Direction;
	EntropyContainer<V> * Entropy;

	bool newF = false,
	         tagACK, tagRST, tagFIN;
	         //tagPSH, tagURG, tagSYN;

	if (pPkt->Protocol() == PKT_PTC_UDP)
	{
		UDPPacket const * udpP = reinterpret_cast<UDPPacket const *>(pPkt);

		srcIP = *(udpP->SrcIPaddr());
		dstIP = *(udpP->DstIPaddr());

		sPort = udpP->SrcPort();
		dPort = udpP->DstPort();

		Proto = "UDP";
	}
	else if (pPkt->Protocol() == PKT_PTC_TCP)
	{
		TCPPacket const * tcpP = reinterpret_cast<TCPPacket const *>(pPkt);

		srcIP = *(tcpP->SrcIPaddr());
		dstIP = *(tcpP->DstIPaddr());

		sPort = tcpP->SrcPort();
		dPort = tcpP->DstPort();

		tagACK = tcpP->Ack();
		tagRST = tcpP->Rst();
		tagFIN = tcpP->Fin();
//		tagPSH = tcpP->Psh();
//		tagSYN = tcpP->Syn();
//		tagURG = tcpP->Urg();

		Proto = "TCP";
	}
	else
		return NULL;

	double TStamp = pPkt->TSseconds() + ( (pPkt->TSmicrosec()) * 0.000001 );

	std::stringstream ssKey1 (std::stringstream::out), ssKey2 (std::stringstream::out);

	ssKey1 << srcIP << "." << sPort << ":" << dstIP << "." << dPort << ":" << Proto;
	ssKey2 << dstIP << "." << dPort << ":" << srcIP << "." << sPort << ":" << Proto;

	const std::string Key1 = ssKey1.str(), Key2 = ssKey2.str();
	std::string Key;

	if (this->Flows.count(Key1) != 0)
	{
		Key = Key1;
		Direction = CS;
	}
	else if (this->Flows.count(Key2) != 0)
	{
		Key = Key2;
		Direction = SC;
	}
	else
	{
		Key = Key1;
		Direction = CS;
		newF = true;
	}

	if (Proto == "UDP")
	{
		if (newF)
			this->NewFlow(Key, Value, TStamp, srcIP, sPort, dstIP, dPort, Proto);
		else
		{
			double LastTimeS = this->LastTS[Key];

			if ((TStamp - LastTimeS) > MAX_UDP_TIMEOUT)
			{
				this->CloseConnection(Key);
				Key = Key1;
				this->NewFlow(Key, Value, TStamp, srcIP, sPort, dstIP, dPort, Proto);
			}
			else
			{
				(this->NumPackets[Key])++;
				this->LastTS[Key] = TStamp;
			}
		}

		Entropy = this->Flows[Key];

		if (this->LastValue.count(Key) != 0)
		{
			D difVal = Value - (this->LastValue[Key]);
			D RoundDif = this->Round<D>(difVal, precision);

			Entropy->AddValue(RoundDif);
			this->WriteEntropy(Key, TStamp, RoundDif, Entropy);
			this->LastValue[Key] = Value;
		}
		else
			this->LastValue.insert(std::pair<std::string, V>(Key, Value));

		(this->TotalUDPpackets)++;
	}

	else if (Proto == "TCP")
	{
		if (newF)
			this->NewFlow(Key, Value, TStamp, srcIP, sPort, dstIP, dPort, Proto);
		else
		{
			double LastTimeS = this->LastTS[Key];

			if ( ((TStamp - LastTimeS) > (2 * MSL)) && 	// Timeout = 2 MSL --- based on RFC 793
				  ( this->HalfClosedCSflows[Key] == RST ||
					this->HalfClosedSCflows[Key] == RST ||
				   (this->HalfClosedCSflows[Key] == CLOSED && this->HalfClosedSCflows[Key] == CLOSED) ) )
			{
				this->CloseConnection(Key);
				Key = Key1;
				this->NewFlow(Key, Value, TStamp, srcIP, sPort, dstIP, dPort, Proto);
			}
			else
			{
				(this->NumPackets[Key])++;
				this->LastTS[Key] = TStamp;
			}
		}

		Entropy = this->Flows[Key];

		if (this->LastValue.count(Key) != 0)
		{
			D difVal = Value - (this->LastValue[Key]);
			D RoundDif = this->Round<D>(difVal, precision);

			Entropy->AddValue(RoundDif);
			this->WriteEntropy(Key, TStamp, RoundDif, Entropy);
			this->LastValue[Key] = Value;
		}
		else
			this->LastValue.insert(std::pair<std::string, V>(Key, Value));

		(this->TotalTCPpackets)++;
		this->UpdateHalfClosedFlows(Key, Direction, tagFIN, tagRST, tagACK);
	}

	return Entropy->GetStats();
}

template<class V> double * FlowAnalyzer<V>::AddValue(const Packet *pPkt, const V &Value)
{
    std::string srcIP, dstIP, Proto;
    int sPort, dPort, Direction;
    EntropyContainer<V> * Entropy;

    bool newF = false,
         tagACK, tagRST, tagFIN;
         //tagPSH, tagURG, tagSYN;

    if (pPkt->Protocol() == PKT_PTC_UDP)
    {
        UDPPacket const * udpP = reinterpret_cast<UDPPacket const *>(pPkt);

        srcIP = *(udpP->SrcIPaddr());
        dstIP = *(udpP->DstIPaddr());

        sPort = udpP->SrcPort();
        dPort = udpP->DstPort();

        Proto = "UDP";
    }
    else if (pPkt->Protocol() == PKT_PTC_TCP)
    {
        TCPPacket const * tcpP = reinterpret_cast<TCPPacket const *>(pPkt);

        srcIP = *(tcpP->SrcIPaddr());
        dstIP = *(tcpP->DstIPaddr());

        sPort = tcpP->SrcPort();
        dPort = tcpP->DstPort();

        tagACK = tcpP->Ack();
//        tagPSH = tcpP->Psh();
        tagRST = tcpP->Rst();
//        tagSYN = tcpP->Syn();
        tagFIN = tcpP->Fin();
//        tagURG = tcpP->Urg();

        Proto = "TCP";
    }
    else
        return NULL;

    double TStamp = pPkt->TSseconds() + ( (pPkt->TSmicrosec()) * 0.000001 );

    std::stringstream ssKey1 (std::stringstream::out), ssKey2 (std::stringstream::out);

    ssKey1 << srcIP << "." << sPort << ":" << dstIP << "." << dPort << ":" << Proto;
    ssKey2 << dstIP << "." << dPort << ":" << srcIP << "." << sPort << ":" << Proto;

    const std::string Key1 = ssKey1.str(), Key2 = ssKey2.str();
    std::string Key;

    if (this->Flows.count(Key1) != 0)
    {
    	Key = Key1;
    	Direction = CS;
    }
    else if (this->Flows.count(Key2) != 0)
    {
    	Key = Key2;
    	Direction = SC;
    }
    else
    {
        Key = Key1;
        Direction = CS;
        newF = true;
    }

    if (Proto == "UDP")
    {
        if (newF)
            this->NewFlow(Key, Value, TStamp, srcIP, sPort, dstIP, dPort, Proto);
        else
        {
            double LastTimeS = this->LastTS[Key];

            if ((TStamp - LastTimeS) > MAX_UDP_TIMEOUT)
            {
                this->CloseConnection(Key);
                Key = Key1;
                this->NewFlow(Key, Value, TStamp, srcIP, sPort, dstIP, dPort, Proto);
            }
            else
            {
                (this->NumPackets[Key])++;
                this->LastTS[Key] = TStamp;
            }
        }

        Entropy = this->Flows[Key];
        Entropy->AddValue(Value);

        (this->TotalUDPpackets)++;

        this->WriteEntropy(Key, TStamp, Value, Entropy);
    }

    else if (Proto == "TCP")
    {
        if (newF)
            this->NewFlow(Key, Value, TStamp, srcIP, sPort, dstIP, dPort, Proto);
        else
        {
            double LastTimeS = this->LastTS[Key];

            if ( ((TStamp - LastTimeS) > (2 * MSL)) && 	// Timeout = 2 MSL --- based on RFC 793
                  ( this->HalfClosedCSflows[Key] == RST ||
                	this->HalfClosedSCflows[Key] == RST ||
                   (this->HalfClosedCSflows[Key] == CLOSED && this->HalfClosedSCflows[Key] == CLOSED) ) )
            {
                this->CloseConnection(Key);
                Key = Key1;
                this->NewFlow(Key, Value, TStamp, srcIP, sPort, dstIP, dPort, Proto);
            }
            else
            {
                (this->NumPackets[Key])++;
                this->LastTS[Key] = TStamp;
            }
        }

        Entropy = this->Flows[Key];
        Entropy->AddValue(Value);

        (this->TotalTCPpackets)++;

        this->WriteEntropy(Key, TStamp, Value, Entropy);

        this->UpdateHalfClosedFlows(Key, Direction, tagFIN, tagRST, tagACK);
    }

    return Entropy->GetStats();
}

template<class V> void FlowAnalyzer<V>::CloseConnection(const std::string &Key)
{
    delete (this->Flows[Key]);
    this->Flows.erase(Key);
    this->HalfClosedCSflows.erase(Key);
    this->HalfClosedSCflows.erase(Key);
    this->NumPackets.erase(Key);
    this->LastTS.erase(Key);
    this->LastValue.erase(Key);
}

template<class V> void FlowAnalyzer<V>::UpdateHalfClosedFlows(const std::string &Key, const int &Direction, const bool &tagFIN, const bool &tagRST, const bool &tagACK)
{
	if (Direction == CS)
	{
		if (tagRST)
			this->HalfClosedCSflows[Key] = RST;

		else
		{
			if ( tagFIN &&
				 this->HalfClosedCSflows[Key] != RST &&
				 this->HalfClosedSCflows[Key] != RST &&
    			 this->HalfClosedCSflows[Key] != CLOSED )
				this->HalfClosedCSflows[Key] = FIN;

			if ( tagACK && this->HalfClosedSCflows[Key] == FIN )
				this->HalfClosedSCflows[Key] = CLOSED;
		}
    }

    else if (Direction == SC)
    {
    	if (tagRST)
			this->HalfClosedSCflows[Key] = RST;

    	else
    	{
    		if ( tagFIN &&
    			 this->HalfClosedCSflows[Key] != RST &&
    			 this->HalfClosedSCflows[Key] != RST &&
    			 this->HalfClosedSCflows[Key] != CLOSED )
				this->HalfClosedSCflows[Key] = FIN;

			if ( tagACK && this->HalfClosedCSflows[Key] == FIN )
				this->HalfClosedCSflows[Key] = CLOSED;
    	}
    }
}

template<class V> void FlowAnalyzer<V>::NewFlow(const std::string &Key, const V &Value, const double &TStamp, const std::string &srcIP, const int &sPort, const std::string &dstIP, const int dPort, const std::string &Proto)
{
    this->Flows.insert(std::pair<std::string, EntropyContainer<V> *>(Key, new EntropyContainer<V>(this->WinSize)));
    this->NumPackets.insert(std::pair<std::string, long>(Key, 1));
    this->LastTS.insert(std::pair<std::string, double>(Key, TStamp));
    this->HalfClosedCSflows.insert(std::pair<std::string, int>(Key, INITIAL));
    this->HalfClosedSCflows.insert(std::pair<std::string, int>(Key, INITIAL));

    if (this->NumFlows.count(Key) == 0)
    	this->NumFlows.insert(std::pair<std::string, int>(Key, 1));
    else
    	(this->NumFlows[Key])++;

    if (Proto == "UDP")
        (this->TotalUDPflows)++;
    else if (Proto == "TCP")
        (this->TotalTCPflows)++;

    time_t rawtime;
    struct tm * timeinfo;

    time ( &rawtime );
    timeinfo = localtime ( &rawtime );

    std::stringstream ss (std::stringstream::out);
    ss << "#\n# Log for flow " << srcIP << ":" << sPort << " -> " << dstIP << ":" << dPort << " " << Proto <<
          "\n#\n# Created on " << asctime (timeinfo) << "#\n" <<
          "# <Timestamp>	<Time elapsed>   <Analyzed value>   <Entropy for a window of " << this->WinSize << " packets>	<Mean of the entropy values>	<Variance of the entropy values>	<MIN of the entropy values>		<MAX of the entropy values>			<Mean of all the values>		<Variance of all the values>	<MIN value in the window>	<MAX value in the window>	<Mean of the MIN values>	<Mean of the MAX values>	<Variance of the MIN values>	<Variance of the MAX values>\n#\n";

    this->WriteNewLog(Key, ss.str());
}

template<> void FlowAnalyzer<std::string>::NewFlow(const std::string &Key, const std::string &Value, const double &TStamp, const std::string &srcIP, const int &sPort, const std::string &dstIP, const int dPort, const std::string &Proto)
{
    this->Flows.insert(std::pair<std::string, EntropyContainer<std::string> *>(Key, new EntropyContainer<std::string>(this->WinSize)));
    this->NumPackets.insert(std::pair<std::string, long>(Key, 1));
    this->LastTS.insert(std::pair<std::string, double>(Key, TStamp));
    this->HalfClosedCSflows.insert(std::pair<std::string, int>(Key, INITIAL));
    this->HalfClosedSCflows.insert(std::pair<std::string, int>(Key, INITIAL));

    if (this->NumFlows.count(Key) == 0)
    	this->NumFlows.insert(std::pair<std::string, int>(Key, 1));
    else
    	(this->NumFlows[Key])++;

    if (Proto == "UDP")
        (this->TotalUDPflows)++;
    else if (Proto == "TCP")
        (this->TotalTCPflows)++;

    time_t rawtime;
    struct tm * timeinfo;

    time ( &rawtime );
    timeinfo = localtime ( &rawtime );

    std::stringstream ss (std::stringstream::out);
    ss << "#\n# Log for flow " << srcIP << ":" << sPort << " -> " << dstIP << ":" << dPort << " " << Proto <<
          "\n#\n# Created on " << asctime (timeinfo) << "#\n" <<
          "# <Timestamp>	<Time elapsed>   <Analyzed value>   <Entropy for a window of " << this->WinSize << " packets>	<Mean of the entropy values>	<Variance of the entropy values>	<MIN of the entropy values>		<MAX of the entropy values>\n#\n";

    this->WriteNewLog(Key, ss.str());
}

template<class V> template<class D> const D FlowAnalyzer<V>::Round(const D &number, const int &precision) const
{
	D Aux, Result;
	int precFactor, intRes;

	if (precision == 0)
		return number;
	else if (precision > 0)
	{
		precFactor = 10;
		for (int i=1; i<precision; i++)
			precFactor *= 10;
	}
	else
		return number;

	Aux = number * (D) precFactor;
	intRes = (int) Aux;

	Result = intRes / (D) precFactor;

    return Result;
}

template<class V> unsigned long FlowAnalyzer<V>::TotUDPflows() const
{
    return this->TotalUDPflows;
}

template<class V> unsigned long FlowAnalyzer<V>::TotUDPpackets() const
{
    return this->TotalUDPpackets;
}

template<class V> unsigned long FlowAnalyzer<V>::TotTCPflows() const
{
    return this->TotalTCPflows;
}

template<class V> unsigned long FlowAnalyzer<V>::TotTCPpackets() const
{
    return this->TotalTCPpackets;
}

template class FlowAnalyzer<int>;
template class FlowAnalyzer<double>;
template class FlowAnalyzer<std::string>;
template const int FlowAnalyzer<int>::Round<int>(const int &number, const int &precision) const;
template const double FlowAnalyzer<double>::Round<double>(const double &number, const int &precision) const;
template double * FlowAnalyzer<int>::AddValueDif<int>(const Packet *pPkt, const int &Value, int precision);
template double * FlowAnalyzer<double>::AddValueDif<double>(const Packet *pPkt, const double &Value, int precision);
