#ifndef FLOWANALYZER_H
#define FLOWANALYZER_H

#ifdef __cplusplus
extern "C"
{
#endif
#include <time.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <string.h>
#ifdef __cplusplus
}
#endif

#include <entropycontainer.h>
#include <direxception.h>
#include <Folder.h>
#include <thresholdsset.h>
#include <map>
#include <string>
#include <fstream>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <pcapwrap.h>
#include <slidingwindow.h>

#ifndef MAX_TIMEOUT
#define MAX_TIMEOUT
#define MAX_UDP_TIMEOUT 64  // (in seconds) based on Claffy et al., "A parameterizable methodology for Internet traffic flow profiling"
#define MSL 120 			// (in seconds) Maximum Segment Life for TCP: 2 minutes based on RFC 793
#endif

#ifndef DIRECTIONS
#define DIRECTIONS
#define CS 1	// client-server direction (client is usually the one who starts the connection)
#define SC 2
#endif

#ifndef TCP_STATE
#define TCP_STATE
#define CLOSED 0
#define RST 1
#define FIN 2
#define INITIAL 3
#endif

template<class V>
class FlowAnalyzer
{
    public:
        FlowAnalyzer(const int &WindowSize);
        FlowAnalyzer(const int &WindowSize, const std::string &LogsPath);
        FlowAnalyzer(const int &WindowSize, const std::string &LogsPath, ThresholdsSet * TSet);
        virtual ~FlowAnalyzer();
        double * AddValue(const Packet *pPkt, const V &Value);
        template<class D> double * AddValueDif(const Packet *pPkt, const D &Value, int precision);
        unsigned long TotTCPflows() const;
        unsigned long TotTCPpackets() const;
        unsigned long TotUDPflows() const;
        unsigned long TotUDPpackets() const;
        void Clear();
    protected:
    private:
        std::map<std::string, EntropyContainer<V> *> Flows;	// < Flow , Entropy Container >
        std::map<std::string, int> HalfClosedCSflows;		// < Flow , state >
        std::map<std::string, int> HalfClosedSCflows;		// < Flow , state >
        std::map<std::string, double> LastTS;				// < Flow , timestamp >
        std::map<std::string, long> NumPackets;       		// < Flow , Number of Packets >
        std::map<std::string, int> NumFlows;				// < Flow , Number of Flows with the same Key >
        std::map<std::string, V> LastValue;					// < Flow, Last value (used by AddValueDif) >
        std::map<std::string, double> FirstTS;
        unsigned long TotalTCPflows;
        unsigned long TotalTCPpackets;
        unsigned long TotalUDPflows;
        unsigned long TotalUDPpackets;
        int WinSize;
        Folder LogsFolder;
        ThresholdsSet * Thresholds;

        void init(const int &WindowSize, const std::string &Path, ThresholdsSet * TSet);
        void dispose();
        void WriteEntropy(const std::string &Key, const double &TStamp, const V &Value, EntropyContainer<V> * Entropy);
        void WriteNewLog(const std::string &Key, const std::string &Msg);
        void WriteExistentLog(const std::string &Key, const std::string &Msg);
        void NewFlow(const std::string &Key, const V &Value, const double &TStamp, const std::string &srcIP, const int &sPort, const std::string &dstIP, const int dPort, const std::string &Proto);
        void CloseConnection(const std::string &Key);
        void UpdateHalfClosedFlows(const std::string &Key, const int &Direction, const bool &tagFIN, const bool &tagRST, const bool &tagACK);
        template <class D> const D Round(const D &number, const int &precision) const;
};

#endif // FLOWANALYZER_H
