#include "hostanalyzer.h"

template<class V> HostAnalyzer<V>::HostAnalyzer(const int &WindowSize)
{
    init(WindowSize, std::string(), 0, false);
}

template<class V> HostAnalyzer<V>::HostAnalyzer(const int &WindowSize, const std::string &LogsPath)
{
    init(WindowSize, LogsPath, 0, true);
}

template<class V> HostAnalyzer<V>::HostAnalyzer(const int &WindowSize, const std::string &LogsPath, ThresholdsSet * TSet)
{
    init(WindowSize, LogsPath, TSet, true);
}

template<class V> HostAnalyzer<V>::HostAnalyzer(const int &WindowSize, const std::string &LogsPath, const bool &PrintLogs)
{
    init(WindowSize, LogsPath, 0, PrintLogs);
}

template<class V> HostAnalyzer<V>::HostAnalyzer(const int &WindowSize, const std::string &LogsPath, ThresholdsSet * TSet, const bool &PrintLogs)
{
    init(WindowSize, LogsPath, TSet, PrintLogs);
}

template<class V> HostAnalyzer<V>::~HostAnalyzer()
{
    this->dispose();
}

template<class V> void HostAnalyzer<V>::Clear()
{
    this->dispose();

    this->Hosts.clear();
    this->LastValue.clear();

    this->TotalHosts = 0;
    this->TotalValues = 0;

    this->Thresholds = 0;

    this->WinSize = 0;
}

template<class V> void HostAnalyzer<V>::dispose()
{
    typename std::map<std::string,EntropyContainer<V> *>::iterator it;

    for(it = this->Hosts.begin(); it != this->Hosts.end(); it++)
        delete (it->second);
}

template<class V> void HostAnalyzer<V>::init(const int &WindowSize, const std::string &Path, ThresholdsSet * TSet, const bool &PrintL)
{
    this->TotalHosts = 0;
    this->TotalValues = 0;
    this->WinSize = WindowSize;

    this->Thresholds = TSet;

    if (PrintL)
    	this->LogsFolder.InitFolder(Path);
    this->PrintLogs = PrintL;
}

template<class V> void HostAnalyzer<V>::SetPrintLogs(const bool &PrintL)
{
	this->PrintLogs = PrintL;
}

template<class V> void HostAnalyzer<V>::SetLogsFolter(const std::string &LogsPath)
{
	this->LogsFolder.InitFolder(LogsPath);
}

template<class V> void HostAnalyzer<V>::WriteEntropy(const std::string &HostIP, const long &TSsec, const long &TSusec, const V &Value, EntropyContainer<V>* Entropy)
{
    std::stringstream ss (std::stringstream::out);

	const double dTS = TSsec + ( TSusec * 0.000001 );
	double dTimeElapsed;

    if (this->FirstTS.count(HostIP) == 0)
		this->FirstTS.insert(std::pair<std::string, double>(HostIP, dTS));

    dTimeElapsed = dTS - (this->FirstTS[HostIP]);

    if (this->Thresholds != 0)
    	ss << TSsec << "." << std::setfill('0') << std::setw(6) << TSusec << "\t" << std::fixed << std::setprecision(6) << dTimeElapsed << "\t" << Value << "\t" << Entropy->EntropyVal() << "\t" << Entropy->EntropyMean() << "\t" << Entropy->EntropyVariance() << "\t" << Entropy->EntropyMin() << "\t" << Entropy->EntropyMax() << "\t" << std::setprecision(2) << Entropy->ValueMean() << "\t" << Entropy->ValueVariance() << "\t" << Entropy->Min() << "\t" << Entropy->Max() << "\t" << Entropy->MinMean() << "\t" << Entropy->MaxMean() << "\t" << Entropy->MinVariance() << "\t" << Entropy->MaxVariance() << "\t" << this->Thresholds->GetDescription(Entropy->EntropyVal()) << std::endl;
    else
    	ss << TSsec << "." << std::setfill('0') << std::setw(6) << TSusec << "\t" << std::fixed << std::setprecision(6) << dTimeElapsed << "\t" << Value << "\t" << Entropy->EntropyVal() << "\t" << Entropy->EntropyMean() << "\t" << Entropy->EntropyVariance() << "\t" << Entropy->EntropyMin() << "\t" << Entropy->EntropyMax() << "\t" << std::setprecision(2) << Entropy->ValueMean() << "\t" << Entropy->ValueVariance() << "\t" << Entropy->Min() << "\t" << Entropy->Max() << "\t" << Entropy->MinMean() << "\t" << Entropy->MaxMean() << "\t" << Entropy->MinVariance() << "\t" << Entropy->MaxVariance() << std::endl;

    this->WriteExistentLog(HostIP, ss.str());
}

template<> void HostAnalyzer<std::string>::WriteEntropy(const std::string &HostIP, const long &TSsec, const long &TSusec, const std::string &Value, EntropyContainer<std::string>* Entropy)
{
    std::stringstream ss (std::stringstream::out);

	const double dTS = TSsec + ( TSusec * 0.000001 );
	double dTimeElapsed;

    if (this->FirstTS.count(HostIP) == 0)
		this->FirstTS.insert(std::pair<std::string, double>(HostIP, dTS));

    dTimeElapsed = dTS - (this->FirstTS[HostIP]);

    if (this->Thresholds != 0)
    	ss << TSsec << "." << std::setfill('0') << std::setw(6) << TSusec << "\t" << std::fixed << std::setprecision(6) << dTimeElapsed << "\t" << Value << "\t" << Entropy->EntropyVal() << "\t" << Entropy->EntropyMean() << "\t" << Entropy->EntropyVariance() << "\t" << Entropy->EntropyMin() << "\t" << Entropy->EntropyMax() << "\t" << this->Thresholds->GetDescription(Entropy->EntropyVal()) << std::endl;
    else
    	ss << TSsec << "." << std::setfill('0') << std::setw(6) << TSusec << "\t" << std::fixed << std::setprecision(6) << dTimeElapsed << "\t" << Value << "\t" << Entropy->EntropyVal() << "\t" << Entropy->EntropyMean() << "\t" << Entropy->EntropyVariance() << "\t" << Entropy->EntropyMin() << "\t" << Entropy->EntropyMax() << std::endl;

    this->WriteExistentLog(HostIP, ss.str());
}

template<class V> void HostAnalyzer<V>::WriteNewLog(const std::string &HostIP, const std::string &Msg) const
{
	if (this->PrintLogs)
	{
		std::string fileName = HostIP + ".log";

		this->LogsFolder.WriteNewFile(fileName, Msg);
	}
}

template<class V> void HostAnalyzer<V>::WriteExistentLog(const std::string &HostIP, const std::string &Msg) const
{
	if (this->PrintLogs)
	{
		std::string fileName = HostIP + ".log";

		this->LogsFolder.WriteExistentFile(fileName, Msg);
	}
}

template<class V> template<class D> double * HostAnalyzer<V>::AddValueDif(const std::string &HostIP, const D &Value, int precision)
{
	return this->AddValueDif(0, 0, HostIP, Value, precision);
}

template<class V> template<class D> double * HostAnalyzer<V>::AddValueDif(const long &TSsec, const long &TSusec, const std::string &HostIP, const D &Value, int precision)
{
	if (this->LastValue.count(HostIP) == 0)
	{
		this->LastValue.insert(std::pair<std::string, D>(HostIP, Value));
		return 0;
	}

	D difVal = Value - (this->LastValue[HostIP]);
	D RoundDif = this->Round<D>(difVal, precision);

	this->LastValue[HostIP] = Value;

	return this->AddValue(TSsec, TSusec, HostIP, RoundDif);
}

template<class V> double * HostAnalyzer<V>::AddValue(const std::string &HostIP, const V &Value)
{
    return this->AddValue(0, 0, HostIP, Value);
}

template<class V> double * HostAnalyzer<V>::AddValue(const long &TSsec, const long &TSusec, const std::string &HostIP, const V &Value)
{
    if (this->Hosts.count(HostIP) == 0)
    {
        this->Hosts.insert(std::pair<std::string,EntropyContainer<V> *>(HostIP,new EntropyContainer<V>(this->WinSize)));
        (this->TotalHosts)++;

        time_t rawtime;
        struct tm * timeinfo;

        time ( &rawtime );
        timeinfo = localtime ( &rawtime );

        std::stringstream ss (std::stringstream::out);
        ss << "#\n# Log for host " << HostIP <<
              "\n#\n# Created on " << asctime (timeinfo) << "#\n" <<
              "# <Timestamp>	<Time elapsed>   <Analyzed value>   <Entropy for a window of " << this->WinSize << " packets>	<Mean of the entropy values>	<Variance of the entropy values>	<MIN of the entropy values>		<MAX of the entropy values>			<Mean of all the values>		<Variance of all the values>	<MIN value in the window>	<MAX value in the window>	<Mean of the MIN values>	<Mean of the MAX values>	<Variance of the MIN values>	<Variance of the MAX values>\n#\n";

        this->WriteNewLog(HostIP, ss.str());
    }

    EntropyContainer<V> * Entropy = this->Hosts[HostIP];
    Entropy->AddValue(Value);

    (this->TotalValues)++;

    this->WriteEntropy(HostIP, TSsec, TSusec, Value, Entropy);

    return Entropy->GetStats();
}

template<> double * HostAnalyzer<std::string>::AddValue(const long &TSsec, const long &TSusec, const std::string &HostIP, const std::string &Value)
{
    if (this->Hosts.count(HostIP) == 0)
    {
        this->Hosts.insert(std::pair<std::string,EntropyContainer<std::string> *>(HostIP,new EntropyContainer<std::string>(this->WinSize)));
        (this->TotalHosts)++;

        time_t rawtime;
        struct tm * timeinfo;

        time ( &rawtime );
        timeinfo = localtime ( &rawtime );

        std::stringstream ss (std::stringstream::out);
        ss << "#\n# Log for host " << HostIP <<
              "\n#\n# Created on " << asctime (timeinfo) << "#\n" <<
              "# <Timestamp>	<Time elapsed>   <Analyzed value>   <Entropy for a window of " << this->WinSize << " packets>	<Mean of the entropy values>	<Variance of the entropy values>	<MIN of the entropy values>		<MAX of the entropy values>\n#\n";

        this->WriteNewLog(HostIP, ss.str());
    }

    EntropyContainer<std::string> * Entropy = this->Hosts[HostIP];
    Entropy->AddValue(Value);

    (this->TotalValues)++;

    this->WriteEntropy(HostIP, TSsec, TSusec, Value, Entropy);

    return Entropy->GetStats();
}

template<class V> template<class D> const D HostAnalyzer<V>::Round(const D &number, const int &precision) const
{
	D Aux, Result;
	int precFactor, intRes;

	if (precision == 0)
		return number;
	else if (precision > 0)
	{
		precFactor = 10;
		for (int i=1; i<precision; i++)
			precFactor *= 10;
	}
	else
		return number;

	Aux = number * (D) precFactor;
	intRes = (int) Aux;

	Result = intRes / (D) precFactor;

    return Result;
}

template<class V> unsigned long HostAnalyzer<V>::TotHosts() const
{
    return this->TotalHosts;
}

template<class V> unsigned long HostAnalyzer<V>::TotValues() const
{
    return this->TotalValues;
}

template class HostAnalyzer<int>;
template class HostAnalyzer<double>;
template class HostAnalyzer<std::string>;
template const int HostAnalyzer<int>::Round<int>(const int &number, const int &precision) const;
template const double HostAnalyzer<double>::Round<double>(const double &number, const int &precision) const;
template double * HostAnalyzer<int>::AddValueDif<int>(const std::string &HostIP, const int &Value, int precision);
template double * HostAnalyzer<double>::AddValueDif<double>(const std::string &HostIP, const double &Value, int precision);
template double * HostAnalyzer<int>::AddValueDif<int>(const long &TSsec, const long &TSusec, const std::string &HostIP, const int &Value, int precision);
template double * HostAnalyzer<double>::AddValueDif<double>(const long &TSsec, const long &TSusec, const std::string &HostIP, const double &Value, int precision);

