#include "hostportanalyzer.h"

template<class V> HostPortAnalyzer<V>::HostPortAnalyzer(const int &WindowSize) {
	init(WindowSize, std::string(), true, 0);
}

template<class V> HostPortAnalyzer<V>::HostPortAnalyzer(const int &WindowSize,
		const std::string &LogsPath) {
	init(WindowSize, LogsPath, true, 0);
}

template<class V> HostPortAnalyzer<V>::HostPortAnalyzer(const int &WindowSize,
		const std::string &LogsPath, const bool &Print) {
	init(WindowSize, LogsPath, Print, 0);
}

template<class V> HostPortAnalyzer<V>::HostPortAnalyzer(const int &WindowSize,
		const std::string &LogsPath, ThresholdsSet * TSet) {
	init(WindowSize, LogsPath, true, TSet);
}

template<class V> HostPortAnalyzer<V>::HostPortAnalyzer(const int &WindowSize,
		const std::string &LogsPath, const bool &Print, ThresholdsSet * TSet) {
	init(WindowSize, LogsPath, Print, TSet);
}

template<class V> HostPortAnalyzer<V>::~HostPortAnalyzer() {
	this->dispose();
}

template<class V> void HostPortAnalyzer<V>::Clear() {
	this->dispose();

	this->Hosts.clear();

	this->TotalHosts = 0;
	this->TotalValues = 0;

	this->Thresholds = 0;

	this->WinSize = 0;
}

template<class V> void HostPortAnalyzer<V>::dispose() {
	typename std::map<std::string, EntropyContainer<V> *>::iterator it;

	for (it = this->Hosts.begin(); it != this->Hosts.end(); it++)
		delete (it->second);
}

template<class V> void HostPortAnalyzer<V>::init(const int &WindowSize,
		const std::string &Path, const bool &Print, ThresholdsSet * TSet) {
	this->TotalHosts = 0;
	this->TotalValues = 0;
	this->WinSize = WindowSize;
	this->PrintOutput = Print;

	this->Thresholds = TSet;

	if (this->PrintOutput)
		this->LogsFolder.InitFolder(Path);
}

template<class V> void HostPortAnalyzer<V>::WriteEntropy(
		const std::string &HostIP, const int &Port, const long &TSsec,
		const long &TSusec, const V &Value, EntropyContainer<V>* Entropy) {
	if (this->PrintOutput) {
		std::stringstream ssKey(std::stringstream::out);
		ssKey << HostIP << ":" << Port;
		std::string Key = ssKey.str();

		std::stringstream ss(std::stringstream::out);

		const double dTS = TSsec + (TSusec * 0.000001);
		double dTimeElapsed;

		if (this->FirstTS.count(Key) == 0)
			this->FirstTS.insert(std::pair<std::string, double>(Key, dTS));

		dTimeElapsed = dTS - (this->FirstTS[Key]);

		if (this->Thresholds != 0)
			ss << TSsec << "." << std::setfill('0') << std::setw(6) << TSusec
					<< "\t" << std::fixed << std::setprecision(6)
					<< dTimeElapsed << "\t" << Value << "\t"
					<< Entropy->EntropyVal() << "\t" << Entropy->EntropyMean()
					<< "\t" << Entropy->EntropyVariance() << "\t"
					<< Entropy->EntropyMin() << "\t" << Entropy->EntropyMax()
					<< "\t" << std::setprecision(2) << Entropy->ValueMean()
					<< "\t" << Entropy->ValueVariance() << "\t"
					<< Entropy->Min() << "\t" << Entropy->Max() << "\t"
					<< Entropy->MinMean() << "\t" << Entropy->MaxMean() << "\t"
					<< Entropy->MinVariance() << "\t" << Entropy->MaxVariance()
					<< "\t" << this->Thresholds->GetDescription(
					Entropy->EntropyVal()) << std::endl;
		else
			ss << TSsec << "." << std::setfill('0') << std::setw(6) << TSusec
					<< "\t" << std::fixed << std::setprecision(6)
					<< dTimeElapsed << "\t" << Value << "\t"
					<< Entropy->EntropyVal() << "\t" << Entropy->EntropyMean()
					<< "\t" << Entropy->EntropyVariance() << "\t"
					<< Entropy->EntropyMin() << "\t" << Entropy->EntropyMax()
					<< "\t" << std::setprecision(2) << Entropy->ValueMean()
					<< "\t" << Entropy->ValueVariance() << "\t"
					<< Entropy->Min() << "\t" << Entropy->Max() << "\t"
					<< Entropy->MinMean() << "\t" << Entropy->MaxMean() << "\t"
					<< Entropy->MinVariance() << "\t" << Entropy->MaxVariance()
					<< std::endl;

		this->WriteExistentLog(HostIP, Port, ss.str());
	}
}

template<> void HostPortAnalyzer<std::string>::WriteEntropy(
		const std::string &HostIP, const int &Port, const long &TSsec,
		const long &TSusec, const std::string &Value, EntropyContainer<
				std::string>* Entropy) {
	if (this->PrintOutput) {
		std::stringstream ssKey(std::stringstream::out);
		ssKey << HostIP << ":" << Port;
		std::string Key = ssKey.str();

		std::stringstream ss(std::stringstream::out);

		const double dTS = TSsec + (TSusec * 0.000001);
		double dTimeElapsed;

		if (this->FirstTS.count(Key) == 0)
			this->FirstTS.insert(std::pair<std::string, double>(Key, dTS));

		dTimeElapsed = dTS - (this->FirstTS[Key]);

		if (this->Thresholds != 0)
			ss << TSsec << "." << std::setfill('0') << std::setw(6) << TSusec
					<< "\t" << std::fixed << std::setprecision(6)
					<< dTimeElapsed << "\t" << Value << "\t"
					<< Entropy->EntropyVal() << "\t" << Entropy->EntropyMean()
					<< "\t" << Entropy->EntropyVariance() << "\t"
					<< Entropy->EntropyMin() << "\t" << Entropy->EntropyMax()
					<< "\t" << this->Thresholds->GetDescription(
					Entropy->EntropyVal()) << std::endl;
		else
			ss << TSsec << "." << std::setfill('0') << std::setw(6) << TSusec
					<< "\t" << std::fixed << std::setprecision(6)
					<< dTimeElapsed << "\t" << Value << "\t"
					<< Entropy->EntropyVal() << "\t" << Entropy->EntropyMean()
					<< "\t" << Entropy->EntropyVariance() << "\t"
					<< Entropy->EntropyMin() << "\t" << Entropy->EntropyMax()
					<< std::endl;

		this->WriteExistentLog(HostIP, Port, ss.str());
	}
}

template<class V> void HostPortAnalyzer<V>::WriteNewLog(
		const std::string &HostIP, const int &Port, const std::string &Msg) const {
	if (this->PrintOutput) {
		std::stringstream fileName(std::stringstream::out);

		fileName << HostIP << "." << Port << ".log";

		this->LogsFolder.WriteNewFile(fileName.str().c_str(), Msg);
	}
}

template<class V> void HostPortAnalyzer<V>::WriteExistentLog(
		const std::string &HostIP, const int &Port, const std::string &Msg) const {
	if (this->PrintOutput) {
		std::stringstream fileName(std::stringstream::out);

		fileName << HostIP << "." << Port << ".log";

		this->LogsFolder.WriteExistentFile(fileName.str().c_str(), Msg);
	}
}

template<class V> template<class D> double * HostPortAnalyzer<V>::AddValueDif(
		const std::string &HostIP, const int &Port, const D &Value,
		int precision) {
	return this->AddValueDif(0, 0, HostIP, Port, Value, precision);
}

template<class V> template<class D> double * HostPortAnalyzer<V>::AddValueDif(
		const long &TSsec, const long &TSusec, const std::string &HostIP,
		const int &Port, const D &Value, int precision) {
	std::stringstream ssKey(std::stringstream::out);
	ssKey << HostIP << ":" << Port;
	std::string Key = ssKey.str();

	if (this->LastValue.count(Key) == 0) {
		this->LastValue.insert(std::pair<std::string, D>(Key, Value));
		return 0;
	}

	D difVal = Value - (this->LastValue[Key]);
	D RoundDif = this->Round<D> (difVal, precision);

	this->LastValue[Key] = Value;

	return this->AddValue(TSsec, TSusec, HostIP, Port, RoundDif);
}

template<class V> double * HostPortAnalyzer<V>::AddValue(
		const std::string &HostIP, const int &Port, const V &Value) {
	return this->AddValue(0, 0, HostIP, Port, Value);
}

template<class V> double * HostPortAnalyzer<V>::AddValue(const long &TSsec,
		const long &TSusec, const std::string &HostIP, const int &Port,
		const V &Value) {
	std::stringstream ssKey(std::stringstream::out);
	ssKey << HostIP << ":" << Port;
	const std::string Key = ssKey.str();

	if (this->Hosts.count(Key) == 0) {
		this->Hosts.insert(std::pair<std::string, EntropyContainer<V> *>(Key,
				new EntropyContainer<V> (this->WinSize)));
		(this->TotalHosts)++;

		time_t rawtime;
		struct tm * timeinfo;

		time(&rawtime);
		timeinfo = localtime(&rawtime);

		std::stringstream ss(std::stringstream::out);
		ss << "#\n# Log for host " << HostIP << " on Port " << Port
				<< "\n#\n# Created on " << asctime(timeinfo) << "#\n"
				<< "# <Timestamp>	<Time elapsed>   <Analyzed value>   <Entropy for a window of "
				<< this->WinSize
				<< " packets>	<Mean of the entropy values>	<Variance of the entropy values>	<MIN of the entropy values>		<MAX of the entropy values>			<Mean of all the values>		<Variance of all the values>	<MIN value in the window>	<MAX value in the window>	<Mean of the MIN values>	<Mean of the MAX values>	<Variance of the MIN values>	<Variance of the MAX values>\n#\n";

		this->WriteNewLog(HostIP, Port, ss.str());
	}

	EntropyContainer<V> * Entropy = this->Hosts[Key];
	Entropy->AddValue(Value);

	(this->TotalValues)++;

	this->WriteEntropy(HostIP, Port, TSsec, TSusec, Value, Entropy);

	return Entropy->GetStats();
}

template<> double * HostPortAnalyzer<std::string>::AddValue(const long &TSsec,
		const long &TSusec, const std::string &HostIP, const int &Port,
		const std::string &Value) {
	std::stringstream ssKey(std::stringstream::out);
	ssKey << HostIP << ":" << Port;
	const std::string Key = ssKey.str();

	if (this->Hosts.count(Key) == 0) {
		this->Hosts.insert(std::pair<std::string,
				EntropyContainer<std::string> *>(Key, new EntropyContainer<
				std::string> (this->WinSize)));
		(this->TotalHosts)++;

		time_t rawtime;
		struct tm * timeinfo;

		time(&rawtime);
		timeinfo = localtime(&rawtime);

		std::stringstream ss(std::stringstream::out);
		ss << "#\n# Log for host " << HostIP << " on Port " << Port
				<< "\n#\n# Created on " << asctime(timeinfo) << "#\n"
				<< "# <Timestamp>	<Time elapsed>   <Analyzed value>   <Entropy for a window of "
				<< this->WinSize
				<< " packets>	<Mean of the entropy values>	<Variance of the entropy values>	<MIN of the entropy values>		<MAX of the entropy values>\n#\n";

		this->WriteNewLog(HostIP, Port, ss.str());
	}

	EntropyContainer<std::string> * Entropy = this->Hosts[Key];
	Entropy->AddValue(Value);

	(this->TotalValues)++;

	this->WriteEntropy(HostIP, Port, TSsec, TSusec, Value, Entropy);

	return Entropy->GetStats();
}

template<class V> void HostPortAnalyzer<V>::ClearHostPort(const std::string &Host, const int &Port) {

	std::stringstream ssKey(std::stringstream::out);
	ssKey << Host << ":" << Port;
	const std::string HostPort = ssKey.str();

	if (this->LastValue.count(HostPort) > 0)
		this->LastValue.erase(HostPort);

	if (this->Hosts.count(HostPort) > 0)
	{
		this->FirstTS.erase(HostPort);
		delete this->Hosts.at(HostPort);
		this->Hosts.erase(HostPort);
	}
}

template<class V> unsigned long HostPortAnalyzer<V>::TotHosts() const {
	return this->TotalHosts;
}

template<class V> unsigned long HostPortAnalyzer<V>::TotValues() const {
	return this->TotalValues;
}

template<class V> template<class D> const D HostPortAnalyzer<V>::Round(
		const D &number, const int &precision) const {
	D Aux, Result;
	int precFactor, intRes;

	if (precision == 0)
		return number;
	else if (precision > 0) {
		precFactor = 10;
		for (int i = 1; i < precision; i++)
			precFactor *= 10;
	} else
		return number;

	Aux = number * (D) precFactor;
	intRes = (int) Aux;

	Result = intRes / (D) precFactor;

	return Result;
}

template class HostPortAnalyzer<int> ;
template class HostPortAnalyzer<double> ;
template class HostPortAnalyzer<std::string> ;
template const int HostPortAnalyzer<int>::Round<int>(const int &number, const int &precision) const;
template const double HostPortAnalyzer<double>::Round<double>(const double &number, const int &precision) const;
template double * HostPortAnalyzer<int>::AddValueDif<int>(const std::string &HostIP, const int &Port, const int &Value, int precision);
template double * HostPortAnalyzer<double>::AddValueDif<double>(const std::string &HostIP, const int &Port, const double &Value, int precision);
template double * HostPortAnalyzer<int>::AddValueDif<int>(const long &TSsec, const long &TSusec, const std::string &HostIP, const int &Port, const int &Value, int precision);
template double * HostPortAnalyzer<double>::AddValueDif<double>(const long &TSsec, const long &TSusec, const std::string &HostIP, const int &Port, const double &Value, int precision);
